# -*- coding: utf-8 -*-
#
# Trespasser Blender Scripts
#
# Copyright (C) 2011, Plüss Roland ( roland@rptd.ch )
# 
# This program is free software; you can redistribute it and/or 
# modify it under the terms of the GNU General Public License 
# as published by the Free Software Foundation; either 
# version 2 of the License, or (at your option) any later 
# version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
# NOTE: For the GPL copy see http://www.gnu.org/licenses/gpl.html
#

import bpy
import os
import re
import math
import time
import struct
import mathutils

from bpy_extras.io_utils import ExportHelper
from mathutils import Vector, Matrix

from .tp_math import transformPosition, convertMatrix, vector_by_matrix, axis_by_matrix, matrixToEuler
from .tp_resources import Instance, Mesh, Armature



class OBJECT_OT_ExportTPM( bpy.types.Operator, ExportHelper ):
	bl_idname = "trespasser.export_tpm"
	bl_label = "TPM (.tpm)"
	__doc__ = """Export as Tresspasser TPM"""
	filename_ext = ".tpm"
	filter_glob = bpy.props.StringProperty( default="*.tpm", options={ 'HIDDEN' } )
	debug_level = bpy.props.EnumProperty( items = (
		( "0", "None", "Output no debug messages." ),
		( "1", "Basic", "Output basic amount of debug messages." ),
		( "2", "Verbose", "Output lots of debug messages." ),
		( "3", "Debug", "Output very large amount of debug messages." ),
		), name = "Logging", description = "Choose amount of logging", default = "1" )
	
	def __init__( self ):
		self.instances = []
		self.meshes = {}
		self.debugLevel = 1
	
	@classmethod
	def poll( cls, context ):
		return context.active_object != None
	
	def execute( self, context ):
		self.debugLevel = int( self.debug_level )
		self.report( { 'INFO' }, "export to '%s'..." % ( self.filepath ) )
		
		self.scaling = context.scene.trespasser_scaling
		self.scalePosition = Matrix( ( (self.scaling,0,0,0), (0,self.scaling,0,0), (0,0,self.scaling,0), (0,0,0,1) ) )
		self.transformScalePosition = self.scalePosition * transformPosition
		
		self.ignoreBones = []
		
		self.export( context )
		
		return { 'FINISHED' }
	
	def export( self, context ):
		self.initExporterObjects( context )
		if not self.checkInitState( context ):
			return False
		f = open( self.filepath, "w" )
		try:
			result = self.safeExport( context, f )
		finally:
			f.close()
		self.printInfos( context )
		return result
	
	def initExporterObjects( self, context ):
		self.initFindResources( context )
		for m in self.meshes.values():
			if m.armature:
				m.armature.ignoreBones = self.ignoreBones
				m.armature.initAddBones()
			m.initAddTextures()
			m.initAddTexCoordSets()
			m.initAddVertices()
			m.initAddEdges()
			m.initAddFaces()
			m.initCalcCornerNormals()
			m.initCalcNormals()
			m.initCalcInfoNumbers()
			m.sortFacesByTexture()
			m.determineExportNames()
	
	def initFindResources( self, context ):
		for object in context.selected_objects:
			if object.type == 'MESH':
				if object.data.name in self.meshes.values():
					mesh = self.meshes[ object.data.name ]
				else:
					mesh = Mesh( object )
					self.meshes[ object.data.name ] = mesh
				self.instances.append( Instance( object, mesh, self.transformScalePosition ) )
				if object.parent and object.parent.type == 'ARMATURE':
					if not mesh.armature:
						mesh.armature = Armature( object.parent )
	
	def checkInitState( self, context ):
		if len( self.meshes ) == 0:
			self.report( { 'INFO', 'ERROR' }, "There is no Mesh selected. Select at last one Mesh" )
			return False
		for mesh in self.meshes.values():
			if mesh.multiFoldMesh:
				self.report( { 'INFO', 'ERROR' }, "Can not Multi-Fold meshes (Edges used by more than 2 faces)." )
				return False
			if mesh.degeneratedFaceCount > 0:
				self.report( { 'INFO', 'ERROR' }, "Degenerated Faces found." )
				return False
		return True
	
	def printInfos( self, context ):
		for mesh in self.meshes.values():
			mesh.printInfos()
			if mesh.armature:
				mesh.armature.printInfos()
		for instance in self.instances:
			instance.printInfos()
	
	def safeExport( self, context, f ):
		if not self.writeHeader( f ):
			return False
		if not self.writeMaterials( f ):
			return False
		if not self.writeMeshes( f ):
			return False
		if not self.writeInstances( f ):
			return False
		return True	
	
	# write header to file
	def writeHeader( self, f ):
		f.write( "fileinfo\n" )
		f.write( "{\n" )
		f.write( " formatversion = 1.0.2\n" )
		f.write( "}\n" )
		return True
	
	# write materials
	def writeMaterials( self, f ):
		if self.debugLevel > 0:
			print( "saving materials..." )
		for mesh in self.meshes.values():
			for texture in mesh.textures:
				f.write( "\n" )
				f.write( "material \"%s\"\n" % texture.exportName )
				f.write( "{\n" )
				f.write( " colormap = \"%s\"\n" % texture.exportBitmap )
				f.write( "}\n" )
		return True
	
	# write models
	def writeMeshes( self, f ):
		if self.debugLevel > 0:
			print( "saving meshes..." )
		uvDummy = [ [ 0, 0 ], [ 1, 0 ], [ 1, 1 ], [ 0, 1 ] ]
		for mesh in self.meshes.values():
			f.write( "\n" )
			f.write( "mesh \"%s\"\n" % mesh.exportName )
			f.write( "{\n" )
			
			# materials
			for texture in mesh.textures:
				f.write( " m = \"%s\"\n" % texture.exportName )
			
			# vertex positions
			for vertex in mesh.vertices:
				wpos = vector_by_matrix( self.transformScalePosition, vertex.vertex.co )
				f.write( " v = (%f,%f,%f)\n" % ( wpos.x, wpos.y, wpos.z ) )
			
			# face texture coordinates
			if len( mesh.texCoordSets ) > 0:
				uvlayer = mesh.texCoordSets[ 0 ].uvlayer
				for face in mesh.faces:
					if len( face.vertices ) == 3 or len( face.vertices ) == 4:
						for uv in uvlayer.data[ face.index ].uv:
							f.write( " t = (%f,%f)\n" % ( uv[ 0 ], uv[ 1 ] ) )
			else:
				for face in mesh.faces:
					if len( face.vertices ) == 3 or len( face.vertices ) == 4:
						for corner in range( len( face.vertices) ):
							f.write( " t = (%f,%f)\n" % ( uvDummy[ corner ][ 0 ], uvDummy[ corner ][ 1 ] ) )
			
			# normals
			for face in mesh.faces:
				if len( face.vertices ) == 3 or len( face.vertices ) == 4:
					for ni in face.normals:
						normal = axis_by_matrix( self.transformScalePosition, mesh.normals[ ni ] )
						f.write( " n = (%f,%f,%f)\n" % ( normal.x, normal.y, normal.z ) )
			
			# faces
			base = 0
			for face in mesh.faces:
				textureIndex = 0
				if face.texture:
					textureIndex = face.face.material_index + 1
				if len( face.vertices ) == 3 or len( face.vertices ) == 4:
					f.write( " f = (%i,%i,%i),(%i,%i,%i),(%i,%i,%i),%i\n" % ( \
						face.vertices[ 0 ] + 1, face.vertices[ 1 ] + 1, face.vertices[ 2 ] + 1, \
						base + 1, base + 2, base + 3, \
						base + 1, base + 2, base + 3,
						textureIndex ) )
					if len( face.vertices ) == 4:
						f.write( " f = (%i,%i,%i),(%i,%i,%i),(%i,%i,%i),%i\n" % ( \
							face.vertices[ 0 ] + 1, face.vertices[ 2 ] + 1, face.vertices[ 3 ] + 1, \
							base + 1, base + 3, base + 4, \
							base + 1, base + 3, base + 4,
							textureIndex ) )
					base = base + len( face.vertices )
			
			# done and done
			f.write( "}\n" )
		return True
	
	# write instances
	def writeInstances( self, f ):
		if self.debugLevel > 0:
			print( "saving instances..." )
		for instance in self.instances:
			f.write( "\n" )
			f.write( "instance \"%s\"\n" % instance.exportName )
			f.write( "{\n" )
			f.write( " mesh = \"%s\"\n" % instance.mesh.exportName )
			f.write( " position = (%f,%f,%f)\n" % ( instance.position.x, instance.position.y, instance.position.z ) )
			f.write( " rotation = (%f,%f,%f)\n" % ( instance.rotation.x, instance.rotation.y, instance.rotation.z ) )
			f.write( " scale = %f\n" % instance.scaling.x )
			f.write( "}\n" )
		return True
