# -*- coding: utf-8 -*-
#
# Trespasser Blender Scripts
#
# Copyright (C) 2011, Plüss Roland ( roland@rptd.ch )
# 
# This program is free software; you can redistribute it and/or 
# modify it under the terms of the GNU General Public License 
# as published by the Free Software Foundation; either 
# version 2 of the License, or (at your option) any later 
# version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
# NOTE: For the GPL copy see http://www.gnu.org/licenses/gpl.html
#

import bpy
import os
import re
import math
import time
import struct
import mathutils

from bpy_extras.io_utils import ExportHelper
from mathutils import Vector, Matrix

from .tp_math import transformPosition, convertMatrix, vector_by_matrix, axis_by_matrix, matrixToEuler
from .tp_resources import Instance, Mesh, Armature



class OBJECT_OT_ExportValuesTXT( bpy.types.Operator, ExportHelper ):
	bl_idname = "trespasser.export_valuestxt"
	bl_label = "Values.txt (.values.txt)"
	__doc__ = """Export as Tresspasser Values.txt"""
	filename_ext = ".values_txt"
	filter_glob = bpy.props.StringProperty( default="*.values_txt", options={ 'HIDDEN' } )
	debug_level = bpy.props.EnumProperty( items = (
		( "0", "None", "Output no debug messages." ),
		( "1", "Basic", "Output basic amount of debug messages." ),
		( "2", "Verbose", "Output lots of debug messages." ),
		( "3", "Debug", "Output very large amount of debug messages." ),
		), name = "Logging", description = "Choose amount of logging", default = "1" )
	
	def __init__( self ):
		self.instances = []
		self.debugLevel = 1
	
	@classmethod
	def poll( cls, context ):
		return context.active_object != None
	
	def execute( self, context ):
		self.filepath = re.sub( "\.values_txt$", ".values.txt", self.filepath )
		
		self.debugLevel = int( self.debug_level )
		self.report( { 'INFO' }, "export to '%s'..." % ( self.filepath ) )
		
		self.scaling = context.scene.trespasser_scaling
		self.scalePosition = Matrix( ( (self.scaling,0,0,0), (0,self.scaling,0,0), (0,0,self.scaling,0), (0,0,0,1) ) )
		self.transformScalePosition = self.scalePosition * transformPosition
		
		self.ignoreBones = []
		
		self.export( context )
		
		return { 'FINISHED' }
	
	def export( self, context ):
		self.initExporterObjects( context )
		if not self.checkInitState( context ):
			return False
		f = open( self.filepath, "w" )
		try:
			result = self.safeExport( context, f )
		finally:
			f.close()
		self.printInfos( context )
		return result
	
	def initExporterObjects( self, context ):
		self.initFindResources( context )
	
	def initFindResources( self, context ):
		for object in context.selected_objects:
			if object.type == 'MESH':
				self.instances.append( Instance( object, None, self.transformScalePosition ) )
	
	def checkInitState( self, context ):
		if len( self.instances ) == 0:
			self.report( { 'INFO', 'ERROR' }, "There is no Mesh selected. Select at last one Mesh" )
			return False
		return True
	
	def printInfos( self, context ):
		for instance in self.instances:
			instance.printInfos()
	
	def safeExport( self, context, f ):
		if not self.writeValues( f ):
			return False
		return True	
	
	# get instance with the given object or None if not found
	def getInstanceWithObject( self, object ):
		for instance in self.instances:
			if instance.object == object:
				return instance
		return None
	
	# write values to file
	def writeValues( self, f ):
		for instance in self.instances:
			f.write( "\n" )
			f.write( "group %s = {\n" % instance.exportName )
			if instance.object.trespasser_meshtype == 'PHYSICS':
				self.writeDefPhysicsValues( f, instance, self.isObjectMoveable( instance.object ) )
			elif instance.object.trespasser_meshtype == 'INVISIBLE':
				self.writeDefInvisibleValues( f, instance, self.isObjectMoveable( instance.object ) )
			elif instance.object.trespasser_meshtype == 'MAGNET':
				parentInstance = None
				if instance.object.parent:
					parentInstance = self.getInstanceWithObject( instance.object.parent )
				self.writeDefMagnetValues( f, instance, parentInstance )
			elif instance.object.trespasser_meshtype == 'TRIGGER':
				self.writeDefTriggerValues( f, instance )
			elif instance.object.trespasser_meshtype == 'EFFECT':
				self.writeDefEffectValues( f, instance )
			elif instance.object.trespasser_meshtype == 'VOCAL':
				self.writeDefVocalValues( f, instance )
			else:
				hasPhysics = False
				hasInvisibles = False
				for instance2 in self.instances:
					if instance2.object.parent == instance.object:
						if instance2.object.trespasser_meshtype == 'PHYSICS':
							hasPhysics = True
						elif instance2.object.trespasser_meshtype == 'INVISIBLE':
							hasInvisibles = True
				self.writeDefMeshValues( f, instance, instance.object.trespasser_tangible and ( hasPhysics or hasInvisibles ), self.isObjectMoveable( instance.object ) )
				if hasPhysics:
					f.write( "    string Type = \"Compound\"\n" )
					counter = 0
					for instance2 in self.instances:
						if instance2.object.trespasser_meshtype == 'PHYSICS' and instance2.object.parent == instance.object:
							f.write( "    string Model%.2i = \"%s\"\n" % ( counter, instance2.exportName ) )
							counter = counter + 1
			
			f.write( "}\n" )
		return True
	
	# determines if an object is moveable ( either the object itself or the parent object )
	def isObjectMoveable( self, object ):
		if object.trespasser_moveable:
			return True
		if object.trespasser_meshtype != 'MESH':
			if object.object.parent:
				return self.isObjectMoveable( object.object.parent )
		return False
	
	# write default mesh values
	def writeDefMeshValues( self, f, instance, tangible, moveable ):
		f.write( "    int ext_GeometryType = 2\n" )
		f.write( "    string Class = \"CInstance\"\n" )
		if tangible:
			f.write( "    bool Tangible = false\n" )
		else:
			f.write( "    bool Tangible = true\n" )
		if moveable:
			f.write( "    bool Moveable = true\n" )
		else:
			f.write( "    bool Moveable = false\n" )
		f.write( "    bool Visible = true\n" )
		f.write( "    bool Shadow = true\n" )
		f.write( "    float Density = %f\n" % instance.object.trespasser_density )
		f.write( "    string SoundMaterial = \"%s\"\n" % instance.object.trespasser_soundmaterial )
	
	# write default physics mesh values
	def writeDefPhysicsValues( self, f, instance, moveable ):
		f.write( "    int ext_GeometryType = 1\n" )
		f.write( "    string Class = \"CInstance\"\n" )
		f.write( "    bool Tangible = true\n" )
		if moveable:
			f.write( "    bool Moveable = true\n" )
		else:
			f.write( "    bool Moveable = false\n" )
		f.write( "    bool Visible = false\n" )
		f.write( "    bool Shadow = false\n" )
		f.write( "    string SoundMaterial = \"%s\"\n" % instance.object.trespasser_soundmaterial )
		f.write( "    float Density = %f\n" % instance.object.trespasser_density )
	
	# write default invisible mesh values
	def writeDefInvisibleValues( self, f, instance, moveable ):
		f.write( "    int ext_GeometryType = 1\n" )
		f.write( "    string Class = \"CInstance\"\n" )
		f.write( "    bool Tangible = true\n" )
		if moveable:
			f.write( "    bool Moveable = true\n" )
		else:
			f.write( "    bool Moveable = false\n" )
		f.write( "    bool Visible = false\n" )
		f.write( "    bool Shadow = false\n" )
		f.write( "    string SoundMaterial = \"%s\"\n" % instance.object.trespasser_soundmaterial )
		f.write( "    float Density = %f\n" % instance.object.trespasser_density )
	
	# write default magnet mesh values
	def writeDefMagnetValues( self, f, instance, masterObject ):
		f.write( "    int ext_GeometryType = 1\n" )
		f.write( "    string Class = \"CMagnet\"\n" )
		f.write( "    float Friction = 1.000000\n" )
		if masterObject:
			f.write( "    string MasterObject = \"%s\"\n" % masterObject.exportName )
		f.write( "    bool ZFree = true\n" )
		f.write( "    bool Frozen = false\n" )
	
	# write default trigger mesh values
	def writeDefTriggerValues( self, f, instance ):
		f.write( "    int ext_GeometryType = 1\n" )
		f.write( "    string Class = \"CLocationTrigger\"\n" )
		f.write( "    bool PlayerEnterTrigger = true\n" )
		f.write( "    int FireCount = 1\n" )
		f.write( "    int AlphaChannel = 0\n" )
		f.write( "    int ProcessStyle = 0\n" )
		f.write( "    group Action00 = {\n" )
		f.write( "        int ActionType = 0\n" )
		f.write( "        string Sample = \"something\"\n" )
		f.write( "    }\n" )
	
	# write default effect mesh values
	def writeDefEffectValues( self, f, instance ):
		f.write( "    int ext_GeometryType = 1\n" )
		f.write( "    string Class = \"CParticles\"\n" )
		f.write( "    int R = 255\n" )
		f.write( "    int G = 195\n" )
		f.write( "    int B = 63\n" )
		f.write( "    bool Solid = false\n" )
		f.write( "    group Min = {\n" )
		f.write( "        float Threshold = 0.600000\n" )
		f.write( "        float Number = 0.000000\n" )
		f.write( "        float Velocity = 2.000000\n" )
		f.write( "        float Lifespan = 0.250000\n" )
		f.write( "        float Size = 0.003000\n" )
		f.write( "        float Length = 0.050000\n" )
		f.write( "        float Gravity = 0.000000\n" )
		f.write( "        float GravityStuck = 0.000000\n" )
		f.write( "        float Spread = 90.000000\n" )
		f.write( "    }\n" )
		f.write( "    group Max = {\n" )
		f.write( "        float Threshold = 1.000000\n" )
		f.write( "        float Number = 30.000000\n" )
		f.write( "        float Velocity = 4.000000\n" )
		f.write( "        float Lifespan = 0.350000\n" )
		f.write( "        float Size = 0.005000\n" )
		f.write( "        float Length = 0.090000\n" )
		f.write( "        float Gravity = 0.000000\n" )
		f.write( "        float GravityStuck = 0.000000\n" )
		f.write( "        float Spread = 30.000000\n" )
		f.write( "    }\n" )
		f.write( "    string A00 = \"Oil Barrel 1\"\n" )
		f.write( "    string A01 = \"Metal Dull 2\"\n" )
		f.write( "    string A02 = \"Pistol\"\n" )
		f.write( "    string A03 = \"Metal Hard 1\"\n" )
	
	# write default vocal mesh values
	def writeDefVocalValues( self, f, instance ):
		f.write( "    int ext_GeometryType = 1\n" )
		f.write( "    string Class = \"AI Command\"\n" )
		f.write( "    int Dinosaur = 6\n" )
		f.write( "    int Vocal = 7\n" )
		f.write( "    string A00 = \"ALBR-ROAR-01\"\n" )
		f.write( "    string A01 = \"ALBR-ROAR-02\"\n" )
		f.write( "    string A02 = \"ALBR-CALL-02\"\n" )
