# -*- coding: utf-8 -*-
#
# Trespasser Blender Scripts
#
# Copyright (C) 2011, Plüss Roland ( roland@rptd.ch )
# 
# This program is free software; you can redistribute it and/or 
# modify it under the terms of the GNU General Public License 
# as published by the Free Software Foundation; either 
# version 2 of the License, or (at your option) any later 
# version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
# NOTE: For the GPL copy see http://www.gnu.org/licenses/gpl.html
#

import bpy
import os
import re
import math
import time
import struct
import mathutils

from bpy_extras.io_utils import ExportHelper
from mathutils import Vector, Matrix

from .tp_math import vector_by_matrix, axis_by_matrix, convertEuler, convertMatrix

TP_RE_UNSAFE_NAME_CHARS = "[^A-Za-z0-9_\-$]+"



class Armature:
	class Bone:
		def __init__( self, index, bone ):
			self.index = index
			self.bone = bone
			self.name = bone.name
			self.parent = None
			self.pos = Vector( ( 0.0, 0.0, 0.0 ) )
			self.rot = Vector( ( 0.0, 0.0, 0.0 ) )
			self.restMat = Matrix( ( ( 1.0, 0.0, 0.0, 0.0 ), ( 0.0, 1.0, 0.0, 0.0 ), ( 0.0, 0.0, 1.0, 0.0 ), ( 0.0, 0.0, 0.0, 1.0 ) ) )
		
		def __repr__( self ):
			return "[%i %s]" % ( self.index, self.name )
		
		def __str__( self ):
			return self.__repr__()
	
	class Keyframe:
		def __init__( self, time, pos, rot ):
			self.time = time
			self.pos = pos
			self.rot = rot
		
		def __repr__( self ):
			return "[time=%i pos=%s rot=%s]" % ( self.time, str( self.pos ), str( self.rot ) )
		
		def __str__( self ):
			return self.__repr__()
	
	class MoveBone:
		def __init__( self, bone ):
			self.bone = bone
			self.times = []
			self.keyframes = []
			self.used = False
	
	class Move:
		def __init__( self, action ):
			self.action = action
			self.name = action.name
	
	# constructor
	def __init__( self, object ):
		self.object = object
		self.armature = object.data
		self.bones = []
		self.ignoreBones = []
	
	# add bones
	def initAddBones( self ):
		# add bones in the correct order. for this we simply add
		# all bones without a parent with the recursive helper method
		for bone in self.armature.bones.values():
			if not bone.parent:
				self.initAddBonesHelper( bone, None )
	
	# add bones helper function
	def initAddBonesHelper( self, bone, parent ):
		# check if the bone has to be ignore
		for f in self.ignoreBones:
			if f.match( bone.name ):
				return
		
		# add bone itself
		boneIndex = len( self.bones )
		newBone = Armature.Bone( boneIndex, bone )
		newBone.parent = parent
		self.bones.append( newBone )
		
		# add all children bones in a depth search fashion
		for child in bone.children:
			self.initAddBonesHelper( child, newBone )
	
	# retrieves named bone or None if not found
	def getBoneNamed( self, name ):
		for bone in self.bones:
			if bone.name == name:
				return bone
		return None
	
	# print infos about the armature
	def printInfos( self ):
		print( "*** Armature ***" )
		print( "name:     %s" % ( self.object.name ) )
		print( "bones:    %i" % ( len( self.bones ) ) )



class Mesh:
	class Texture:
		def __init__( self, index, material, textureSlot ):
			self.index = index
			self.material = material
			self.name = material.name
			self.exportName = re.sub( TP_RE_UNSAFE_NAME_CHARS, "_", material.name )
			self.exportBitmap = "something.bmp"
			self.textureSlot = textureSlot
			self.image = None
			self.width = 256
			self.height = 256
			if textureSlot and textureSlot.texture and textureSlot.texture.type == 'IMAGE':
				self.image = textureSlot.texture.image
				if self.image:
					size = self.image.size
					if size[ 0 ] > 0:
						self.width = size[ 0 ]
					if size[ 1 ] > 0:
						self.height = size[ 1 ]
		
		def __repr__( self ):
			return "[%s]" % ( self.name )
		
		def __str__( self ):
			return self.__repr__()
	
	class TexCoordSet:
		def __init__( self, index, uvlayer ):
			self.index = index
			self.uvlayer = uvlayer
			self.name = uvlayer.name
		
		def __repr__( self ):
			return "[%s]" % ( self.name )
		
		def __str__( self ):
			return self.__repr__()
	
	class Weight:
		def __init__( self, bone, weight ):
			self.bone = bone
			self.weight = weight
		
		def equals( self, other ):
			return ( self.bone == other.bone ) and ( abs( self.weight - other.weight ) < 0.001 )
		
		def __repr__( self ):
			return "[%s=%g]" % ( self.bone, self.weight )
		
		def __str__( self ):
			return self.__repr__()
	
	class Weights:
		def __init__( self ):
			self.weights = []
			self.index = 0
			
		def equals( self, other ):
			if len( self.weights ) != len( other.weights ):
				return False
			for i in range( len( self.weights ) ):
				if not self.weights[ i ].equals( other.weights[ i ] ):
					return False
			return True
			
		def normalize( self ):
			weightSum = 0.0
			for weight in self.weights:
				weightSum = weightSum + weight.weight
			if weightSum < 0.001:
				self.weights = []
			else:
				factor = 1.0 / weightSum
				for weight in self.weights:
					weight.weight = weight.weight * factor
		
		def __repr__( self ):
			return self.weights.__repr__()
		
		def __str__( self ):
			return self.__repr__()
	
	class Vertex:
		def __init__( self, index, vertex ):
			self.vertex = vertex
			self.edges = []
			self.faces = []
			self.index = index
			self.weights = Mesh.Weights()
		
		def __repr__( self ):
			return "[%i: edges(%s) faces(%s)]" % ( self.index, str( self.edges ), str( self.faces ) )
		
		def __str__( self ):
			return self.__repr__()
		
		def findEdge( self, edges, v2 ):
			for edge in self.edges:
				if edges[ edge ].vertices[ 1 ] == v2:
					return edge
			return -1
		
		def getHardEdgeCount( self, edges ):
			count = 0
			for edge in self.edges:
				if edges[ edge ].hard:
					count = count + 1
			return count
	
	class Edge:
		def __init__( self, index, v1, v2 ):
			self.vertices = [ v1, v2 ]
			self.faces = [ -1, -1 ]
			self.index = index
			self.hard = False
		
		def __repr__( self ):
			return "[%i: vertices=(%i,%i) hard=%i faces=(%s)]" % ( self.index, self.vertices[ 0 ], self.vertices[ 1 ], self.hard, str( self.faces ) )
		
		def __str__( self ):
			return self.__repr__()
	
	class Face:
		def __init__( self, index, face ):
			self.face = face
			self.vertices = [ face.vertices[ 0 ], face.vertices[ 1 ], face.vertices[ 2 ] ]
			self.normals = [ -1, -1, -1 ]
			self.tangents = [ -1, -1, -1 ]
			self.edges = [ -1, -1, -1 ]
			self.index = index
			self.texture = None
			if len( face.vertices ) == 4:
				self.vertices.append( face.vertices[ 3 ] )
				self.normals.append( -1 )
				self.tangents.append( -1 )
				self.edges.append( -1 )
			self.faceNormal = Vector()
		
		def __repr__( self ):
			return "[%i: vertices(%s) normals(%s) tangents(%s) edges(%s)]" % ( self.index, str( self.vertices ), str( self.normals ), str( self.tangents ), str( self.edges ) )
		
		def __str__( self ):
			return self.__repr__()
		
		def findCorner( self, vertex ):
			for index in range( len( self.vertices ) ):
				if self.vertices[index] == vertex:
					return index
			return -1
		
		def setNormalFor( self, vertex, normal ):
			for index in range( len( self.vertices ) ):
				if self.vertices[ index ] == vertex:
					self.normals[ index ] = normal
		
		def setTangentFor( self, vertex, tangent ):
			for index in range( len( self.vertices ) ):
				if self.vertices[ index ] == vertex:
					self.tangents[ index ] = tangent
	
	# constructor
	def __init__( self, object ):
		self.object = object
		self.mesh = object.data
		self.armature = None
		self.name = object.data.name
		self.exportName = re.sub( TP_RE_UNSAFE_NAME_CHARS, "_", object.data.name )
		self.weights = []
		self.vertices = []
		self.normals = []
		self.edges = []
		self.faces = []
		self.textures = []
		self.texCoordSets = []
		self.triCount = 0
		self.quadCount = 0
		self.edgeCount = 0
		self.normalCount = 0
		self.tangentCount = 0
		self.maxBoneWeightCount = 0
		self.multiFoldMesh = False
		self.degeneratedFaceCount = 0
	
	# add textures
	def initAddTextures( self ):
		for material in self.mesh.materials:
			textureSlot = None
			if len( material.texture_slots ) > 0:
				textureSlot = material.texture_slots[ 0 ]
			self.textures.append( Mesh.Texture( len( self.textures ), material, textureSlot ) )
	
	# add texture coordinates sets
	def initAddTexCoordSets( self ):
		if len( self.mesh.uv_textures ) == 1:
			self.texCoordSets.append( Mesh.TexCoordSet( 0, self.mesh.uv_textures[ 0 ] ) )
		else:
			for uvlayer in self.mesh.uv_textures:
				if uvlayer.name == "default":
					self.texCoordSets.append( Mesh.TexCoordSet( 0, uvlayer ) )
			for uvlayer in self.mesh.uv_textures:
				if uvlayer.name != "default":
					self.texCoordSets.append( Mesh.TexCoordSet( len( self.texCoordSets ), uvlayer ) )
	
	# add vertices
	def initAddVertices( self ):
		# add vertices
		for vertex in self.mesh.vertices:
			newVertex = Mesh.Vertex( vertex.index, vertex )
			self.vertices.append( newVertex )
			# add weights only if there is an armature
			if self.armature:
				# first add all weights as found in the vertex
				for group in vertex.groups:
					bone = self.armature.getBoneNamed( self.object.vertex_groups[ group.group ].name )
					if bone and group.weight > 0.001:
						newVertex.weights.weights.append( Mesh.Weight( bone, group.weight ) )
				# then normalize the weights
				newVertex.weights.normalize()
		# group weights together eliminating duplicates
		for vertex in self.vertices:
			vertexWeights = vertex.weights
			if len( vertexWeights.weights ) > 0:
				for weights in self.weights:
					if vertexWeights.equals( weights ):
						vertex.weights = weights
						vertexWeights = None
						break
				if vertexWeights:
					vertexWeights.index = len( self.weights )
					self.weights.append( vertexWeights )
		# determine the maximum number of bones used in all weights
		for weights in self.weights:
			if len( weights.weights ) > self.maxBoneWeightCount:
				self.maxBoneWeightCount = len( weights.weights )
	
	# add edges
	def initAddEdges( self ):
		for edge in self.mesh.edges:
			newEdge = Mesh.Edge( edge.index, edge.vertices[ 0 ], edge.vertices[ 1 ] )
			newEdge.crease = edge.crease
			newEdge.sharp = edge.use_edge_sharp
			if newEdge.sharp or newEdge.crease > 0.5:
				newEdge.hard = True
			self.edges.append( newEdge )
			self.vertices[ newEdge.vertices[ 0 ] ].edges.append( newEdge.index )
			self.vertices[ newEdge.vertices[ 1 ] ].edges.append( newEdge.index )
	
	# add faces
	def initAddFaces( self ):
		textureCount = len( self.textures )
		for face in self.mesh.faces:
			newFace = Mesh.Face( face.index, face )
			self.faces.append( newFace )
			# face normal
			newFace.faceNormal = face.normal
			# assign texture if possible
			if face.material_index >= 0 and face.material_index < textureCount:
				newFace.texture = self.textures[ face.material_index ]
			# gather vertices and indices
			vo = []
			vi = newFace.vertices
			for e in vi:
				vo.append( self.vertices[ e ] )
			# corners
			count = len( vi )
			for c in range( count ):
				self.initProcessFaceCorner( vi, vo, c, ( c + 1 ) % count, newFace.index )
	
	# init face corner
	def initProcessFaceCorner( self, vi, vo, c1, c2, fi ):
		vo[ c1 ].faces.append( fi )
		edge = vo[ c1 ].findEdge( self.edges, vi[ c2 ] )
		if edge == -1:
			edge = vo[ c2 ].findEdge( self.edges, vi[ c1 ] )
		theEdge = self.edges[ edge ]
		if theEdge.faces[ 0 ] == -1:
			theEdge.faces[ 0 ] = fi
		elif theEdge.faces[ 1 ] == -1:
			theEdge.faces[ 1 ] = fi
		else:
			self.multiFoldMesh = True
		self.faces[ fi ].edges[ c1 ] = edge
	
	# calculate corner normals
	def initCalcCornerNormals( self ):
		self.normalCount = 0
		for vertex in self.vertices:
			hardEdgeCount = vertex.getHardEdgeCount( self.edges )
			if hardEdgeCount == 0:
				for face in vertex.faces:
					self.faces[ face ].setNormalFor( vertex.index, self.normalCount )
				self.normalCount = self.normalCount + 1
			else:
				changed = True
				while changed:
					changed = False
					for index in vertex.faces:
						face = self.faces[ index ]
						c = face.findCorner( vertex.index )
						if face.normals[ c ] == -1:
							edge = self.edges[ face.edges[ c ] ]
							if edge.faces[ 0 ] == face.index:
								neighbor = edge.faces[ 1 ]
							else:
								neighbor = edge.faces[ 0 ]
							if edge.hard or neighbor == -1:
								face.normals[c] = self.normalCount
								self.normalCount = self.normalCount + 1
								changed = True
							else:
								neighborFace = self.faces[ neighbor ]
								neighborCorner = neighborFace.findCorner( vertex.index ) 
								if neighborFace.normals[ neighborCorner ] != -1:
									face.normals[ c ] = neighborFace.normals[ neighborCorner ]
									changed = True
		# ensure no crap normals exist even if things look shit afterwards
		for face in self.faces:
			for normal in range( len( face.normals ) ):
				if face.normals[ normal ] == -1:
					print( "WARNING! -1 normal found at face %i" % ( face ) )
					face.normals[ normal ] = self.normalCount
					self.normalCount = self.normalCount + 1
	
	# calculates the actual normals
	def initCalcNormals( self ):
		for i in range( self.normalCount ):
			self.normals.append( Vector() )
		for face in self.faces:
			for ni in face.normals:
				self.normals[ ni ] = self.normals[ ni ] + face.faceNormal
		for i in range( self.normalCount ):
			length = self.normals[ i ].length
			if length < 0.0001:
				self.normals[ i ] = Vector( 0, 0, 1 )
			else:
				self.normals[ i ] = self.normals[ i ] / length
	
	# calculate informational numbers
	def initCalcInfoNumbers( self ):
		for face in self.faces:
			if len( face.vertices ) == 3:
				self.triCount = self.triCount + 1
			elif len( face.vertices ) == 4:
				self.quadCount = self.quadCount + 1
			else:
				self.degeneratedFaceCount = self.degeneratedFaceCount + 1
		self.edgeCount = len( self.edges )
		
	# sort faces by texture
	def sortFacesByTexture( self ):
		faces = []
		for texture in self.textures:
			for face in self.faces:
				if face.texture == texture:
					faces.append( face )
		self.faces = faces
	
	# determine export names for bitmaps of textures
	def determineExportNames( self ):
		for texture in self.textures:
			exportname = texture.material.trespasser_exportname
			if exportname and exportname != "":
				texture.exportName = "%s.%s" % ( self.exportName, exportname )
			else:
				texture.exportName = "%s.%s" % ( self.exportName, texture.exportName )
			
			bitmapname = texture.material.trespasser_bitmapname
			if bitmapname:
				texture.exportBitmap = bitmapname
			else:
				texture.exportBitmap = "something.bmp"
	
	# print infos
	def printInfos( self ):
		print( "*** Mesh ***" )
		print( "name:         %s" % ( self.name ) )
		print( "exportName:   %s" % ( self.exportName ) )
		print( "textures:     %i" % ( len( self.textures ) ) )
		print( "texCoordSets: %i" % ( len( self.texCoordSets ) ) )
		print( "vertices:     %i" % ( len( self.vertices ) ) )
		print( "faces:        %i" % ( len( self.faces ) ) )
		print( "- tris:       %i" % ( self.triCount ) )
		print( "- quads:      %i" % ( self.quadCount ) )
		print( "edges:        %i" % ( len( self.edges ) ) )
		print( "normals:      %i" % ( self.normalCount ) )
		print( "weights:      %i" % ( len( self.weights ) ) )
		print( "maxBoneWeightCount: %i" % ( self.maxBoneWeightCount ) )



class Instance:
	# constructor
	def __init__( self, object, mesh, matrix ):
		self.object = object
		self.mesh = mesh
		self.name = object.name
		self.exportName = re.sub( TP_RE_UNSAFE_NAME_CHARS, "_", object.name )
		
		(loc,rot,scale) = object.matrix_local.decompose()
		self.position = vector_by_matrix( matrix, loc )
		self.rotation = convertEuler( rot )
		self.scaling = vector_by_matrix( matrix, scale )
		self.scaling = Vector( [ math.fabs( self.scaling.x ), math.fabs( self.scaling.y ), math.fabs( self.scaling.z ) ] )
		self.matrix = object.matrix_local
	
	# print infos
	def printInfos( self ):
		print( "*** Instance ***" )
		print( "name:       %s" % self.name )
		print( "exportName: %s" % self.exportName )
		if self.mesh:
			print( "mesh:       %s" % self.mesh.exportName )
